package net.pingex.discordbot;

import sx.blah.discord.api.EventSubscriber;
import sx.blah.discord.handle.impl.events.MessageReceivedEvent;
import sx.blah.discord.util.DiscordException;
import sx.blah.discord.util.HTTP429Exception;
import sx.blah.discord.util.MissingPermissionsException;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.logging.Logger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * This class parses and dispatch Discord commands
 * @version 0.1-dev
 * @author Raphael "Pingex" NAAS
 */
public class CommandDispatcher
{
	private Logger logger;

	/**
	 * Contains all available commands, built using `rebuildCommandList()`
	 */
	private HashMap<String, InvokableMethod> commandList;

	/**
	 * Basic Constructor, automatically rebuilds command list from the modules registry
	 */
	public CommandDispatcher()
	{
		logger = Logger.getLogger(this.getClass().getName());
		rebuildCommandList();
	}

	/**
	 * Fired when receiving a message
	 * @param event Event from Discord API
	 */
	@EventSubscriber
	public void onMessageReceivedEvent(MessageReceivedEvent event)
	{
		Matcher m = Pattern.compile("^!(\\w+):(\\w+)(?: (.*))?$").matcher(event.getMessage().getContent()); // TODO: Custom command prefix
		if(!m.matches()) return; // We don't need to go further if it's not even a command

		String module = m.group(1);
		String command = m.group(2);
		String fullCommand = module + ":" + command;
		String[] args = (m.group(3) != null) ? m.group(3).split(" ") : null;

		logger.info("Command invoked (" + event.getMessage().getAuthor().getName() + "): " + fullCommand);

		try
		{
			if(commandList.containsKey(fullCommand))
			{
				String answer = (String) commandList.get(fullCommand).invoke(new Object[]{args}); // Bricolage
				if(answer != null) event.getMessage().reply(answer);
			}
			else
				event.getMessage().reply("Unknown command");
		} catch (MissingPermissionsException | HTTP429Exception | DiscordException e)
		{
			logger.warning("Couldn't reply to command: " + e.getMessage());
		} catch (IllegalArgumentException | IllegalAccessException | InvocationTargetException e)
		{
			logger.severe("Couldn't call target method: " + e.getMessage());
		}
	}

	/**
	 * Rebuilds command list using the module registry.
	 * - It looks for methods with @Command annotation inside class with @Controllable annotation
	 * - Check his prototype (should be `method(String[]):String`)
	 * - Add usable methods in his local registry, using InvokableMethod
	 * @see InvokableMethod
	 * @see ModulesRegistry
	 */
	public void rebuildCommandList()
	{
		logger.info("Rebuilding command list...");

		ArrayList<AbstractModule> registry = ModulesRegistry.getRegistry();
		commandList = new HashMap<>();

		for(AbstractModule i : registry)
			if(i.getClass().isAnnotationPresent(Controllable.class))
				for(Method j : i.getClass().getDeclaredMethods())
					if(j.isAnnotationPresent(Command.class))
					{
						String id = i.getClass().getAnnotation(Controllable.class).name() + ":" + j.getName();
						logger.info("Found " + id);

						if(j.getParameterCount() == 1 && j.getParameterTypes()[0] == String[].class && j.getReturnType() == String.class)
							commandList.put(id, new InvokableMethod(j, i));
						else
							logger.warning(id + ": incorrect function prototype, thus won't be added to the command list.");
					}

		logger.info("... Done");
	}
}
